
/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
     nor the names of its contributors may be used to endorse or promote
     products derived from this Software without specific prior written
     permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

package AccordionTreeDrawer;

import java.util.*;
import java.awt.*;
import java.io.*;
import gl4java.*;

import AccordionDrawer.*;
import BitmapFont.*;
/**
 * An abstract class representing a multi-scale rectangular grid on which 
 * a tree consisting of horizontal and vertical line segments is drawn
 *
 * Fills in the fields of the abstract class AccordionDrawer based on
 * information from the Tree class.
 * 
 * @author  Tamara Munzner, Li Zhang
 * @version 
 * @see     AccordionDrawer.AccordionDrawer
 * @see     TreeJuxtaposer.AccordionTreeDrawerFinal
 * @see     AccordionDrawer.Tree
 * @see     AccordionDrawer.TreeNode
 * @see     AccordionDrawer.TreeEdge
 * @see     AccordionDrawer.GridCell
 */

//public class AccordionTreeDrawer extends AccordionDrawer {
public abstract class AccordionTreeDrawer extends AccordionDrawer {
    String fname;
    protected float pointoffset[] = new float[2];
    protected int labeloffset[] = new int[2];
    protected int labelbuffer[] = new int[2];

    /**
     * framebuffer pixels from last mouseover label drawn
     */
    protected float pixels[] = new float[2];
	protected float interactionboxPixels[] = new float[2];

    protected float[][] pal;
    protected float drawAngle;
    /**
     * GL rasterpos of last mouseover label drawn
     */
    protected double drawnlabelpos[] = new double[2];
	protected int interactionBoxPos[] = new int[2];
    /** 
     * width and height of last mouseover label drawn
     */
    protected int drawnlabelsize[] = new int[2];
	protected int interactionBoxSize[] = new int[2];

    //how the labels are drawn
    // dealing with variable font size
    public int minFontHeight = 10;
    public int maxFontHeight = 14;
    public int popupFontHeight = 14;

    protected float forestShrinkFactor;
    protected float pickoffset[] = new float[2];
    public int pickFuzz; 
    //    static String fontFile = "BitmapFont/OlduvaiFont";
    static String fontFile = "/OlduvaiFont";
    static BitmapFontsFamily bff = new BitmapFontsFamily(fontFile);
    static boolean fontInitialized = false;

    public Tree tree;
    protected TreeNode rootNode;
    protected TreeEdge rootEdge;
    public static boolean fullyQualified;
    
    private TreeNode[] leaf;

    // benchmarking 
    public static int countDrawnFrame;
    public static int countDrawnScene;
    public static int cumFrameTime;

	protected Color objectColor;
	public AccordionTreeDrawer(Tree t, int w, int h) {
	super(w,h);
	expandleaves = 1;
	dimbrite = true;
	horiz = true;
	objectColor = new Color(0.0f, 0.0f, 0.0f, 1.0f);
	labeloffset[X] = 2;
	labeloffset[Y] = 4;
	labelbuffer[X] = 10;
	labelbuffer[Y] = 1;
	pointoffset[X] = .1f;
	pointoffset[Y] = .52f;
	pickoffset[0] = .1f;
	pickoffset[1] = .1f;
	pickFuzz = 5; // in pixels
	forestShrinkFactor = -0.3f;
	if(!fontInitialized) {
	    initializeFont();
	    fontInitialized = true;
	}

	tree = t;
	tree.setNodeFontSize(maxFontHeight, minFontHeight);
	rootNode = tree.getRoot();
	initCells();
//	buildGrids();
//	attachEdges();
	flashGeom = tree.getRoot();
	flashGeomOld = null;
	doBox = false;
    drawAngle = 30.0f;

    }

	/**
	 * clear the objects initiated in ATD
	 * @see TreeJuxtaposer.TreeJuxtaposer.deleteTree
	 * @see AccordionDrawer.close
	 */

		public void shutdown(){
			super.shutdown();      
			tree=null;
			rootNode=null;
			rootEdge=null;
			fname=null;	 		
			pointoffset=null;
			labeloffset = null;
			labelbuffer = null;		
			pixels = null;
			interactionboxPixels = null;
			drawnlabelpos = null;	
			drawnlabelsize =null;
		 }
 
		 protected void finalize() throws Throwable {
		 
					 try {
						 shutdown();
					 }
					 finally {
						System.out.println("finally clean ATD");
						 super.finalize();     
					 }
		 }


    /**
     * Initialize variable font family for tree labels. fontFamily[i]
     * is the font should be used when there enough space for font
     * with size i+minFontSize. The reason this is stored in an array
     * is to speed up the look up time. 
     * 
     * @author Li Zhang
     * @see TreeNode.drawInCell
     */
    private void initializeFont() {
	// bitmap fonts
	try {
	    bff.read();
	}catch(IOException e) {
	    System.out.println("cannot produce fonts");
	    return;
	}
	System.out.println("bitmap fonts initialized"+" "+bff.nofonts);
    }

    /**
     * Initialization done after graphics context has been established.
     * @author Tamara Munzner
     */
    public void init() {
	super.init();
	bff.setupGL(gl);
    }

	// get the maxline for some node given children have set minlines
	private int getMaxLineX(ArrayList children)
	{
		int maxLine = bottomSize[X];
		Iterator iter = children.iterator();
		while (iter.hasNext())
		{
			int currNodeMinLine = ((TreeNode)iter.next()).getCell().getMinLine(X);
			if (currNodeMinLine < maxLine)
				maxLine = currNodeMinLine; 
		}
		return maxLine;
	}
	
	// update the children minlines to the maxline of the parent
	private void updateMinLineX(ArrayList children, int newMinLine)
	{
		Iterator iter = children.iterator();
		while (iter.hasNext())
		{
			((TreeNode)iter.next()).getCell().setMinLine(newMinLine, X);
		}
	}
	
	// build all splitLines
	private void buildCells()
	{
		int leafCount = 0;
		Vector leafVector = new Vector();
		int currY = 0; // for positioning leaves in splitLine Y
		for(TreeNode n = tree.posorderStartNode; n!=null; n=n.posorderNext)
		{
			GridCell c = new GridCell(this);
			n.setCell(c);
			c.setCellGeom(n);
			n.computedFrame = -1;
			if (n.isLeaf())
			{  // leaf
				splitLine[X].addCell(c, bottomSize[X]-1);
				splitLine[Y].addCell(c, currY++);
				leafVector.add(n);
				leafCount++;
			}
			else
			{ // internal node
				n.addEdge(new TreeEdge(n, false), Y); // vertical edge
				int maxLineX = getMaxLineX(n.children);
				updateMinLineX(n.children, maxLineX);
				splitLine[X].addCell(c, maxLineX);
				int minLineY = n.getLeftmostLeaf().getCell().getMinLine(Y);
				int maxLineY = n.getRightmostLeaf().getCell().getMaxLine(Y); 
				splitLine[Y].addCell(c, minLineY, maxLineY);
			}
			if (!n.isRoot())
			{
				n.addEdge(new TreeEdge(n, true), X); // horizontal edge
			}
			else
			{
				rootNode = n;
				rootEdge = n.getEdge(Y);
				rootCell = c;
			}
		}
	}

    /**
     * Initializes level 0 GridCells based on Tree tree, assign tree
     * nodes to them. Set {@link #maxUsed maxUsed[]} in directions X
     * and Y.
     *
     *
     * @author   Tamara Munzner, Serdar Tasiran
     * 
     *
     * @see      AccordionDrawer.GridCell
     * @see      AccordionDrawer.Tree
     * @see AccordionDrawer.TreeNode */

    public void initCells() {
	int numLeaves = tree.getLeafCount();
	int treeHeight = tree.getHeight();

	bottomSize[0] = treeHeight; 
	bottomSize[1] = numLeaves;
	initSplitLines();
	splitLine[X].setMinStuckValue(0.01f);

	gridDepth = new int[2];
	double logBase2 = Math.log(2);
	for (int xy = 0; xy < 2; xy++)
	{
		gridDepth[xy] = (int)Math.ceil(Math.log(bottomSize[xy]) / logBase2);
		gridDepth[xy] = gridDepth[xy] <= 0 ? 1 : gridDepth[xy];
	}
	System.out.println("gridDepth: " +gridDepth[0] + " " +gridDepth[1] +
		" rows:numLeaves " + numLeaves + " cols:treeHeight " + treeHeight +
		" total nodes " + tree.getTotalNodeCount() + " treeNum " + tree.getKey());

	buildCells();

	System.out.println("root cell " + rootNode.getCell());
    }

	// x and y are pixel positions
    public CellGeom pickAttached(int x, int y)
    {
		GridCell pickedCell = pickCell(x, y);
		if (pickedCell != null)
			return pickedCell.getCellGeom();
		return null;
    }

    /** 
     * Create box enclosing the subtree beneath input node/edge. The
     * resulting box is the nearest GridLine boundaries to that point,
     * which are cached so that redraws to erase the box can happen
     * after the lines change position.
     * @param cg TreeNode to act upon, passed in as CellGeom from
     *  AccordionDrawer (where TreeNodes are not known)
     * @see IntersectionBox
     * @author Tamara Munzner */
    public InteractionBox makeBox(CellGeom cg)
    {
		if (null == cg)
			return null;
		TreeNode tn;
		if (cg instanceof TreeNode)
		    tn = (TreeNode)cg;
		else if (cg instanceof TreeEdge) 
	    	tn = ((TreeEdge)cg).node;
	    else
	    	return null; // not a node or an edge?
		TreeNode minLeaf = tn.leftmostLeaf;
		TreeNode maxLeaf = tn.rightmostLeaf;
		int[] minLine = {tn.getCell().getMinLine(X), minLeaf.getCell().getMinLine(Y)};
		// maxLine[X] could use any leaf to get rightmost splitLine
		int[] maxLine = {minLeaf.getCell().getMaxLine(X), maxLeaf.getCell().getMaxLine(Y)};
	
		return new InteractionBox(minLine, maxLine, cg, tn.getCell().drawer);
    }



    /**
     * On mouseover, flash the object under the mouse and draw its
     * label at maximum size.
     *
     * @author   Tamara Munzner
     * 
     *
     * @see      AccordionDrawer.GridCell
     * @see      AccordionDrawer.Tree
     * @see      AccordionDrawer.TreeNode
     */
    public void doFlash() 
    {
		if (noflash ||
				(flashGeom == flashGeomOld &&
					flashBox == flashBoxOld) ||
			null == glj ||
			!glj.gljMakeCurrent()) 
	    return;
	    
		doingFlash=true;
		Color objCol = (null == flashCol) ? objectColor : flashCol;
		if (doBox && flashGeom != null) {
	    	flashBox = makeBox(flashGeom);
		}
	
		// order matters: 1) undraw old box 2) dump old pixels 
		//		  3) save new pixels 4) draw box
		if (doBox && flashBoxOld != null) 
			flashBoxOld.undraw();

		boolean drawlabelsreal = drawlabels;
		drawlabels = false;
//System.out.println("flashGeomOld: " + flashGeomOld);
		if (null != flashGeomOld) 
		{
		    Color oldColor;
		    ArrayList temp = getColorsForGridCell(flashGeomOld.getCell(), flashGeomOld.getKey());
		    if (temp == null || temp.size() < 1) oldColor = null; else oldColor = (Color)temp.get(0); 
		    float oldPlane = (null == oldColor) ? objplane : hiliteplane - .004f*oldColor.getAlpha();
	    	oldColor = (null == oldColor) ? objectColor : oldColor;
	    	for (int xy = 0; xy<2; xy++) 
	    	{
				TreeEdge te = ((TreeNode)flashGeomOld).getEdge(xy);
				if (null != te) 
				{
					
					double x=drawnlabelpos[0];
					double y=drawnlabelpos[1];				
			    	String name = te.node.getName();
			    	
			    	//gl.glDisable(GLEnum.GL_DEPTH_TEST);
					
					if(this.drawDirection == UP)
					{
						
						y=drawnlabelpos[1];
			   		    for(int i=0; i<name.length(); i++)
			       		{
					        gl.glRasterPos3d(x,y, getLabelplane());
					        gl.glDrawPixels(drawnlabelsize[0],drawnlabelsize[1], GLEnum.GL_RGB, GLEnum.GL_FLOAT, pal[i]);
		   	    		}
		   			}
			   		else if(this.drawDirection==DOWN)
			 		{
						y=drawnlabelpos[1];			
						for(int i=0; i<name.length(); i++)
				   		{
							gl.glRasterPos3d(x,y, getLabelplane());
							gl.glDrawPixels(drawnlabelsize[0],drawnlabelsize[1], GLEnum.GL_RGB, GLEnum.GL_FLOAT, pal[i]);			

							if(y>getWinMax(Y))
								y=getWinMax(Y)-1;
			   			}
		   			}
					
					
		   	//	else
		   	//	{
		   			if( flag ) // don't draw pixels in the first frame
				//	if( !continueFrame )
						{
//							// if you are not in continue frame, you are still in
//							// startframe-doublebuffer-put-all-in-the-backbuffer-mode
//							// you have to switch to the front buffer
//							//gl.glDrawBuffer(GL_FRONT);
//							//glc.setDoubleBuffered(false);

		  				//gl.glEnable(gl.GL_DEPTH_TEST);
	     				//gl.glDepthFunc(gl.GL_LEQUAL);
		   			    gl.glRasterPos3d(x,y, getLabelplane());
            			gl.glDrawPixels(drawnlabelsize[0],drawnlabelsize[1], GLEnum.GL_RGB, GLEnum.GL_FLOAT, pixels);
						//gl.glDisable(gl.GL_DEPTH_TEST);            			
		   				}					
			 //  	}
				
			//	gl.glEnable(gl.GL_DEPTH_TEST);
			//	gl.glDepthFunc(gl.GL_LEQUAL);
//				System.out.println("oldColor: " + oldColor);
			    te.drawInCell(oldColor, oldPlane);
				// go back in old rendering state
		//		gl.glDisable(GLEnum.GL_DEPTH_TEST);
			} // end if !te
	   	} // end for loop
	} // end if (null != flashGeomOld) 
	
	if (null != flashGeom) 
	{
	    // we don't actually need newcol, we just need to know if
	    // it's non-null to decide whether to paint the object in
	    // the frontmost highlight plane or the rear object plane.
	 	Color newCol;
	 	ArrayList temp = getColorsForGridCell(flashGeom.getCell(), flashGeom.getKey());
	 	if (temp == null || temp.size() < 1) newCol = null; else newCol = (Color)temp.get(0);
	    float newPlane = (null == newCol) ? objplane : hiliteplane - .004f*newCol.getAlpha();

	    // only draw the label for the node. edges just inherit
	    // the identical name from the node, we don't want to draw
	    // it more than once because we're doing the xor trick.
	    for (int xy = 0; xy<2; xy++) 
	    {
			TreeEdge te = ((TreeNode)flashGeom).getEdge(xy);
			if (null != te) 
			{ 
				//gl.glEnable(gl.GL_DEPTH_TEST);
				//gl.glDepthFunc(gl.GL_ALWAYS);
			    te.drawLabelBig(flashX, flashY);
				//gl.glDisable(gl.GL_DEPTH_TEST); 
				te.drawInCell(objCol, newPlane);
				
			}
	    }
	}
	if (doBox && flashBox != null)
	{
		//System.out.println("flashbox draw " + flashBox); 
	    flashBox.draw(rubberbandColor, flashBoxWidth, interactionplane);
	}
	    
	drawlabels = drawlabelsreal;
	if (doBox) flashBoxOld = flashBox;
	flashGeomOld = flashGeom;
	flashXOld = flashX;
	flashYOld = flashY;
	drawEnd();
	doingFlash=false;
    }

    /** 
     * Decide what GridCells to enqueue for drawing based on tree
     * topology. We want to add parents and children of current
     * CellGeom. We should find the GridCell corresponding to the
     * higher-level TreeEdge associated with a TreeNode, not the
     * level0 GridCell of the TreeNode itself.
     * 
     * @author Tamara Munzner
     */
   	// TODO: decide whether or not this function is useful anymore
   	//  - we don't enqueue cells for drawing anymore
    public void drawNextFancy(CellGeom g) {
//	TreeNode tn;
//	if (g instanceof TreeEdge) {
//	    tn = ((TreeEdge)g).node;
//	} else if (g instanceof TreeNode) {
//	    tn = (TreeNode)g;
//	} else {
//	    return;
//	} 
//	TreeNode par = tn.parent();
//	TreeLeafGridCell c;
//	if (null != par && null != par.edges[X]) {
//	    c = (TreeLeafGridCell)par.edges[X].getCell();
//	    c.possiblyEnqueue();
//	} 
//	if (null != tn.children) {
//	    Iterator iter = tn.children.iterator();
//	    while (iter.hasNext()) {
//		TreeNode tn2 = (TreeNode) iter.next();
//		if (null != tn2.edges[X]) {
//		    c = (TreeLeafGridCell)tn2.edges[X].getCell();
//		    c.possiblyEnqueue();
//		}
//	    }
//	}
    }

	protected void drawPreNewFrame() {
	countDrawnFrame = 0;
	countDrawnScene = 0;
	cumFrameTime = 0;
    }

	protected void drawPreContFrame() {
	countDrawnFrame = 0;
    }

	protected void drawPostScene() {
	drawPostFrame();
//	commented so now isn't needed, if needed, pass it properly, not a state machine!
//	if (dumpstats) {
//	    stats.println(" scene "+countDrawnScene+" frametimeCum "+cumFrameTime+" sceneTIME "+(now-continueStart));
//	    //stats.flush();
//	}
    }

	protected void drawPostFrame() {
	long now = System.currentTimeMillis();			
	countDrawnScene += countDrawnFrame;
	long frameTime = now-dynamicStart;
	cumFrameTime += frameTime;
	if (dumpstats) stats.print(" df "+countDrawnFrame+" frameTIME "+frameTime);
    }

	protected void drawBruteForce() {
	drawPreNewFrame();
	drawBruteForceRecurse(rootNode);
	drawPostScene();
    }

    /** 
     * Draw a tree edge, recurse to draw its children.
     * Ignores highlighting, just for benchmarking.
     * @author Tamara Munzner
     */
   	// never called
    private void drawBruteForceRecurse(TreeNode n) {
    	System.out.println("DBFR");
	if (checktime && (0 == countDrawnFrame%100)) {
	    now = System.currentTimeMillis();			
	    if (now-dynamicStart > dynamicTime) {
		return;
	    }
	}
	n.drawInCell(objectColor, objplane);
	Iterator iter = n.children.iterator();
	while (iter.hasNext()) {
	    TreeNode child = (TreeNode)iter.next();
	    if (null != child) drawBruteForceRecurse(child);
	}
    }



    public Tree getTree() { return tree;}
    public Canvas getCanvas() { return this;}
    public TreeNode getNodeByKey(int key) { return tree.getNodeByKey(key);}
    public TreeNode getNodeByName(String name) { return tree.getNodeByName(name);}

    public int getLabelBuffer(int xy) {return labelbuffer[xy];}
    public void setLabelBuffer(int buffer) {
	if (buffer < 1) buffer = 1;
	labelbuffer[X] = buffer;
	labelbuffer[Y] = buffer;
	requestRedraw();
    }
    public void setLabelBuffer(int buffer, int xy) {
	if (buffer < 1) buffer = 1;
	labelbuffer[xy] = buffer;
	requestRedraw();
    }
    public void increaseLabelBuffer(int xy) {
	setLabelBuffer(labelbuffer[xy]+1, xy);
    }
    public void increaseLabelBuffer() {
	setLabelBuffer(labelbuffer[X]+1);
    }
    public void decreaseLabelBuffer(int xy) {
	setLabelBuffer(labelbuffer[xy]-1, xy);
    }
    public void decreaseLabelBuffer() {
	setLabelBuffer(labelbuffer[X]-1);
    }

    public int getMaxFontHeight() {return maxFontHeight;}
    public void setMaxFontHeight(int fontheight) {
	if (fontheight < 1) fontheight = 1;
	maxFontHeight = fontheight;
	tree.setNodeFontSize(maxFontHeight, minFontHeight);
	requestRedraw();
    }
    public void increaseMaxFontHeight() {
	setMaxFontHeight(maxFontHeight+1);
    }
    public void decreaseMaxFontHeight() {
	setMaxFontHeight(maxFontHeight-1);
    }

    public int getMinFontHeight() {return minFontHeight;}
    public void setMinFontHeight(int fontheight) {
	if (fontheight < 1) fontheight = 1;
	minFontHeight = fontheight;
	tree.setNodeFontSize(maxFontHeight, minFontHeight);
	requestRedraw();
    }
    public void increaseMinFontHeight() {
	setMinFontHeight(minFontHeight+1);
    }
    public void decreaseMinFontHeight() {
	setMinFontHeight(minFontHeight-1);
    }

	public void startNewFrameAccordionXDrawer()
	{
		drawnLabels.clear();
	}

	public Color getObjectColor() { return objectColor; }
	public void setObjectColor(Color objectColor) { this.objectColor = objectColor;}

	/* (non-Javadoc)
	 * @see AccordionDrawer.AccordionDrawer#subpixelDraw(AccordionDrawer.GridCell)
	 */
	public void subpixelDraw(GridCell c) {
		// TODO Auto-generated method stub, needed to match accordionDrawer abstract class
	
	}

   public void setDrawAngle(float angle){
   	drawAngle = angle;
   	requestRedraw();
   }

   public float getDrawAngle(){
   	return drawAngle;
   }
    /*
    public static void parse(String[] args) {
	fname = args[0];
    }
    public Color getColorForRange(int objmin, int objmax) {
	return fooColor;
    }
    public static void main(String[] args) {
	try {
	    parse(args);
	    Tree t = new Tree();
	    t.load(fname);
	    //	    t.printLeaves();
	    AccordionTreeDrawer atd = new AccordionTreeDrawer(t, 300, 300);
	    Frame mainFrame = new Frame("AccordionTreeDrawer " + fname);

	    mainFrame.add("Center", atd.canvas);
	    mainFrame.setSize(300,300);
	    mainFrame.pack();
	    mainFrame.setVisible(true);
	    fooColor = new Color(1.0f, 1.0f, 1.0f);
	} catch (IOException  e) {
	    System.out.println("io troubles when parsing file " + fname);
	}
    }
    */
    
	// given some pixel position, return the gridCell
	public GridCell pickCell(int x, int y)
	{
		double xPos = s2w(x, X);
		double yPos = s2w(y, Y);
		double xFuzz = s2w(pickFuzz, X);
		double yFuzz = s2w(pickFuzz, Y);
		if (xPos - xFuzz < splitLine[X].getMinStuckValue() ||
			xPos + xFuzz > splitLine[X].getMaxStuckValue() ||
			yPos - yFuzz < splitLine[Y].getMinStuckValue() ||
			yPos + yFuzz > splitLine[Y].getMaxStuckValue())
		return null; // no cells outside the stuck positions
//		System.out.println("\npicking: " + xPos + " " + yPos);
		return rootNode.pickCell(xPos, yPos, xFuzz, yFuzz);
	}
	
//	public abstract void drawGroups();
	
	/* (non-Javadoc)
	 * @see AccordionDrawer.AccordionDrawer#drawSpecial()
	 */
	public boolean drawSpecial() {
//		splitLine[Y].drawTicks();
		// check to see if layout is correct
		
		// ~5 seconds for 190k nodes
//		splitLine[Y].drawAllRanges();

//		// ~15 seconds for 190k nodes
//		for (int i = 0; i < tree.nodes.size(); i++)
//		{
//			TreeNode tn = (TreeNode)tree.nodes.get(i);
//			tn.drawInCell(null, interactionplane);
////			System.out.println("Drawing: " + tn.getCell());
//		}

		doFrameMove();
//		drawGroups();
		int xy = splitLine[Y].isHorizontal() ? X : Y;
		double pixelSize = s2w(1.0, xy);
		long now = System.currentTimeMillis();
		while (now - dynamicStart < dynamicTime && ToDrawQ.size() > 0)
//			while (ToDrawQ.size() > 0)
			{
				if (!(ToDrawQ.get(0) instanceof RangeInTree))
					System.out.println("what am I in the todrawQ: " + ToDrawQ.get(0).getClass().getName());
			RangeInTree r = (RangeInTree) ToDrawQ.get(0);
			if (r.atd != this)
			{
				System.out.println("Wrong ATD: " + r + " " + this);
//				RangeInTree r2 = r.getSplitLineRange(Y); 
				r.atd.splitLine[Y].drawRange(r.min-1, r.max, pixelSize);
			}
			else
			{
//				System.out.println("DrawRange: " + (r.min-1) + " " + r.max);
//				RangeInTree r2 = r.getSplitLineRange(Y);
				splitLine[Y].drawRange(r.min-1, r.max, pixelSize);
			}
			ToDrawQ.remove(0);
			now = System.currentTimeMillis();
		}
//		System.out.println()
//		//requestRedraw();
//
		drawPostFrame();
//		//gl.glFlush();
//
//		//System.out.println(" bla " + keepMoving + " " + ToDrawQ.size());
		if (keepMoving) {
			requestRedraw();
			drawPostFrame();
		} else if (ToDrawQ.size() > 0) {
			keepDrawing = true;
			continueFrame();
		} else {
			//keepDrawing = false;
			if (colorgrid || drawBackground)
				splitLine[Y].drawBackground(null);
			drawPostScene();
		}
		return true;
	}
	/* (non-Javadoc)
	 * @see AccordionDrawer.AccordionDrawer#drawRange(int, int)
	 */
	// min and max are splitLines indexed in Y, surround leaf cells
	// it's known that the distance between min and max is no more than a pixel
	//  - get the min/max cells, use tree.leaf array
	//   - min is the min index into tree.leaf
	//   - max is the max index into tree.leaf
	//  - find parents of all cells (binary recursion, if min.parent != max.parent, recurse on half)
	// cache drawn frame in each cell so we stop at already drawn ancestor
	public void drawRange(int min, int max)
	{
		if (min == max)
			return; // no split range
		TreeNode minChild = tree.getLeaf(min+1);
		TreeNode maxChild = tree.getLeaf(max);
		if (minChild == null || maxChild == null)
			System.out.println("Error: minChild or maxChild is null: " + minChild + " " + maxChild);
		if (minChild.parent == maxChild.parent)
		{ // common leaves on parent, draw one (base case for recursion)
			minChild.drawInCell(getColorsForRange(minChild.key, maxChild.key), objplane);
			minChild.getCell().setDrawnFrame(frameNum);
			minChild.parent.drawAscend(frameNum, objplane);
		}
		else
		{ // uncommon parent for range, so find all the parents that are in the range
			int mid = (min+max+1)/2;
			TreeNode midChild = tree.getLeaf(mid);
			if (min == mid || minChild.parent == midChild.parent)
			{ // draw leaf between min and mid
				minChild.drawInCell(getColorsForRange(minChild.key, midChild.key), objplane);
				minChild.getCell().setDrawnFrame(frameNum);
				minChild.parent.drawAscend(frameNum, objplane);
			}
			else
			{ // uncommon parent for [min, mid] range, recurse left
				drawRange(min, mid);
				midChild = tree.getLeaf(mid+1);
				if (mid == max || maxChild.parent == midChild.parent)
				{ // draw leaf between mid and max
					maxChild.drawInCell(getColorsForRange(midChild.key, maxChild.key), objplane);
					maxChild.getCell().setDrawnFrame(frameNum);
					maxChild.parent.drawAscend(frameNum, objplane);
				}
				else
				{ // uncommon parent for [mid+1, max] range, recurse right
					drawRange(mid, max);
				}
			}
			
		}
		
//		TreeNode tn = rootNode;
//		GridCell c = tn.getCell();
//		if (c.getMinLine(Y) >= min && c.getMaxLine(Y) <= max)
//		{
//			
//		}
//		Iterator iter = tn.children.iterator();
//		while (iter.hasNext())
//		while (tn.leftmostLeaf.getCell().getMinLine(Y) < min)
	}
	
	public void addRangeToDrawQueue(RangeInTree r)
	{
		if (r != null)
			ToDrawQ.add(r.getSplitLineRange(Y));
	}
	
	public abstract ArrayList getColorsForRange(int min, int max);
	
	public void drawDensity( int minObj, int maxObj)
	{ // only applies to sequences
	}
    
	/**
	 * @return
	 */
	public TreeNode getLeaf(int leafNum) {
		return leaf[leafNum];
	}

};
