
/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
     nor the names of its contributors may be used to endorse or promote
     products derived from this Software without specific prior written
     permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

package AccordionTreeDrawer;

import java.awt.Color;
import java.util.*;
import gl4java.*;
import gl4java.GLFunc;

import AccordionDrawer.*;

/**
 * A class representing an edge of a (phylognenetic) tree
 * 
 * A TreeEdge represents a horizontal or vertical line segment used in drawing a tree.
 *
 * <Pre>
 *     a--- firstChild 
 *     |               
 *     |               
 *     ---  .          
 * n   |    .          
 * o---|c   .	       
 * d   ---             
 * e   |               
 *     |               
 *     b--- lastChild  
 * 
 * </Pre>
 *
 * In this figure (node,c), (a,b), (firstChild,a) and (lastChild,b) are line segments
 * Each of them is represented by a TreeEdge
 * (node,c) corresponds to a horizontal TreeEdge associated with TreeNode node
 * (a,b) corresponds to a vertical TreeEdge associated with TreeNode node
 * 
 * @author  Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
 * @version 
 * @see     AccordionDrawer.Tree
 * @see     AccordionDrawer.TreeNode
 * @see     AccordionDrawer.GridCell
 */

public class TreeEdge {

    public TreeNode node;

    /**
    * Whether this TreeEdge is horizontal
    */ 
    public boolean horiz;


    public TreeEdge(TreeNode n, boolean horizontal)
    {
		node = n;
		horiz = horizontal; // from me to parent I'm horizontal
    }
    
//    public TreeNode getNode()
//    {
//    	return node;
//    }

	private GridCell getCell()
	{
		return node.getCell();
	}

	public AccordionTreeDrawer getDrawer()
	{
		return (AccordionTreeDrawer)getCell().drawer;
	}

//	/**
//	 * clean the tree edge
//	 * @see TreeNode.close
//	 *
//	 */
//	public void close(){
//			node=null;
//	}
    
//	protected void finalize() throws Throwable
//	{	 
//		try
//		{
//			close();
//		}
//		finally
//		{
//			super.finalize();
//		}
//	}
 

//    public int getKey() { return node.getKey(); }
//    public String getName() { return node.getName(); } // never used?
//    public GridCell getCell() { return node.getCell();}
//	public void setCell(GridCell c) { node.setCell(c);}
    
    // these will return splitLine indices
    public int getMin() { return node.getMin(); }
    public int getMax() { return node.getMax(); }

	public void setPositions(float[] start, float[] end, int direction)
	{
		AccordionTreeDrawer atd = getDrawer();
		int X = AccordionDrawer.X, Y = AccordionDrawer.Y;
		GridCell c = getCell();
		double[] minStuck = {atd.getSplitLine(Y).getMinStuckValue(),
			atd.getSplitLine(Y).getMinStuckValue()};
		double[] rangeSize = {
			atd.getSplitLine(X).getMaxStuckValue() - minStuck[X],
			atd.getSplitLine(Y).getMaxStuckValue() - minStuck[Y]};
		if (horiz)
		{
			start[X] = (float)c.getMin(X); // X is over the range
			end[X] = (float)c.getMax(X);
			start[Y] = end[Y] = (float)node.getMidY();
		}
		else
		{
			start[X] = end[X] = (float)c.getMax(X);
			start[Y] = (float)node.getMinY();
			end[Y] = (float)node.getMaxY();
		}

		// draw in reverse direction
		if (direction == AccordionDrawer.RIGHT)
		{
			start[X] = (float)(rangeSize[X] - start[X] + minStuck[X]);
			end[X] = (float)(rangeSize[X] - end[X] + minStuck[X]);
		}
		//TODO: fix drawing in different directions
//		//dawn in up direction
//		else if (direction == AccordionDrawer.UP)
//		{
//			float temp= start[0];
//			start[0]= start[1];
//			start[1] = 1- temp;
//			temp= end[0];
//			end[0]= end[1];
//			end[1] =1- temp;
//		}
//		// draw in down direction
//		else if (direction == AccordionDrawer.DOWN)
//		{
//			float temp= start[1];
//			start[1]= start[0];
//			start[0] =1- temp;
//			temp= end[1];
//			end[1]= end[0];
//			end[0] =1- temp;
//		}
	}
    
    /**
     * Draws this TreeEdge inside the GridCell to which it is
     * attached, and its label if appropriate.
     *
     * Find font size to use for drawing the label, by checking for
     * occlusions. Don't draw it all if it's occluded even at the
     * smallest size.
     * 
     * @author   Tamara Munzner, Li Zhang
     * 
     * @param    col The color to draw this node in
     * @param    plane Z depth at which to draw label
     *
     * @see      AccordionDrawer.GridCell
     */

    public void drawInCell(Color col, double plane)
    {
//    	System.out.println("Drawing: " + node.getCell());
    	int X = horiz ? 0 : 1;
    	int Y = horiz ? 1 : 0;
    AccordionTreeDrawer atd = (AccordionTreeDrawer)getCell().drawer;
	if (col == null)
		col = atd.objectColor;
	if (atd.nogeoms) return;
	AccordionTreeDrawer.countDrawnFrame++;
	GLFunc gl = (GLFunc) atd.getGraphicsStuff();
	
	gl.glEnable(gl.GL_DEPTH_TEST);
	gl.glDepthFunc(gl.GL_LEQUAL);
	
	if (atd.dimcolors || atd.dimbrite) {
	    float rgbcol[] = new float[3];
	    col.getRGBColorComponents(rgbcol);
	    float hsbcol[] = Color.RGBtoHSB((int)(rgbcol[0]*255), 
			 (int)(rgbcol[1]*255), 
			 (int)(rgbcol[2]*255), (new float[3]));
	    float howdim;
	    // H stays the same

	    // if started out greyscale, don't crank up S
	    if (hsbcol[1] > 0f && atd.dimcolors) {
		// depends on size of your subtended subtree screen area
		InteractionBox b = atd.makeBox(node); // box around subtree
		if (null == b) System.out.println("interaction box null");
		// howdim base: one pixel 0, fullscreen 1.0
		howdim = 
			(float)(b.getMax(AccordionDrawer.Y)-b.getMin(AccordionDrawer.Y))/
				(float)atd.getWinMax(1); //screen coords
		// small dynamic range:, 0%=.5, 20% and up =1.0
		howdim = .5f+40*howdim;
		howdim = (howdim > 1.0) ? 1.0f : howdim;
		hsbcol[1] = (float)howdim;
	    }

	    if (atd.dimbrite && col == atd.getObjectColor()) {
		// depends on depth in tree
		int treeheight = atd.tree.getHeight();
		// howdim ranges from 0 at root to 1.0 at leaf
		howdim = (node.height-1.0f)/(treeheight-1.0f);
		// compress the range
		hsbcol[2] = (float)(0.2+howdim*.7);
	    }
	    Color newrgb = Color.getHSBColor(hsbcol[0],hsbcol[1],hsbcol[2]);
	    newrgb.getRGBColorComponents(rgbcol);
	    gl.glColor3f(rgbcol[0], rgbcol[1], rgbcol[2]);
	    //System.out.println("H "+hsbcol[0]+" S "+hsbcol[1]+" B "+hsbcol[2]);
	    //System.out.println("dim " + howdim + " x " + (1.0-howdim*.8) + " th " + treeheight + " nh " + node.height);
	} else {
	    atd.setColorGL(col);
	}
	// draw the line
	float thick = atd.linethickness;
	if (col != atd.getObjectColor() && !atd.isDoingFlash()) thick+=2;
	float[] posStart = new float[2], posEnd = new float[2];
	setPositions(posStart, posEnd, atd.drawDirection);
//	if (horiz)
//		System.out.println(node + " Set positions: Y:" + posEnd[1]);
	gl.glLineWidth(thick);
	gl.glBegin(GLEnum.GL_LINES);
	gl.glVertex3d(posStart[0], posStart[1], plane);
	gl.glVertex3d(posEnd[0], posEnd[1], plane);
	gl.glEnd();
	
	if (horiz)
	{ // draw the dot at the end of the horizontal section
		gl.glPointSize((float)(thick+2.0f));
		gl.glBegin(GLEnum.GL_POINTS);
		gl.glVertex3d(posEnd[0], posEnd[1], plane);
		gl.glEnd();
	}

	if (horiz && atd.drawlabels && node.label.length() > 0) {
	    // first, check smallest font, to see if anything at all
	    // can fit. if it does, then do binary search to find
	    // largest possible box that fits. after we know which box
	    // size to use, draw it.

	    // this algorithm means drawing order will have a radical
	    // effect on how labels look!

	if(node.isInteriorNode())
	{
		int fontSize = node.getFontSize();
//		  fontSize=fontSize<=((AccordionTreeDrawer)d).minFontHeight?
//		  ((AccordionTreeDrawer)d).minFontHeight:fontSize;
		LabelBox lb = makeLabelBox(fontSize, -1, -1, posStart, posEnd);
		boolean intersects = intersectLabelBox(lb);
		if (intersects) return; // nothing will fit
		int fitsheight = fontSize;
		LabelBox fits = lb;
		gl.glEnable(gl.GL_DEPTH_TEST);
		gl.glDepthFunc(gl.GL_LEQUAL);
		drawLabelBox(fits, fitsheight, false);
		gl.glDisable(gl.GL_DEPTH_TEST);
	}
	else
	{
		int min = ((AccordionTreeDrawer)atd).minFontHeight;
		LabelBox lb = makeLabelBox(min, -1, -1, posStart, posEnd);
		boolean intersects = intersectLabelBox(lb);
		if (intersects) return; // nothing will fit
		int fitsheight = min;
		LabelBox fits = lb;
	  // we should set min to one less than the real bound. new
	  // real bound is minFontSize+1 (because we already tested
	  // minFontSize), so leave min set to current value of
	  // (minFontSize+1)-1 = minFontSize
		int max = ((AccordionTreeDrawer)atd).maxFontHeight+1;
		int mid;
		while (max - min > 1)
		{
			mid = (int)((min+max)/2.0);
			lb = makeLabelBox(mid, -1, -1, posStart, posEnd);
			intersects = intersectLabelBox(lb);
			if (intersects) max = mid;
			else min = mid;
		}
		gl.glEnable(gl.GL_DEPTH_TEST);
		gl.glDepthFunc(gl.GL_LEQUAL);
		drawLabelBox(fits, fitsheight, false);
		gl.glDisable(gl.GL_DEPTH_TEST);
	}

  }
  
  gl.glDisable(gl.GL_DEPTH_TEST);
  
  }
    
    /**
     * Draw label of this TreeEdge at maximum size (intended for
     * mouseover highlighting).
     * 
     * @author   Tamara Munzner
     */

    public void drawLabelBig(int x, int y) {
	AccordionTreeDrawer drawer = (AccordionTreeDrawer)getCell().drawer;
	if (!horiz || drawer.nogeoms || node.label.length() < 1) return;
	int fontheight = drawer.popupFontHeight;
	float[] posStart = new float[2], posEnd = new float[2];
	setPositions(posStart, posEnd, drawer.drawDirection);
	LabelBox lb = makeLabelBox(fontheight, x, y, posStart, posEnd);
	drawLabelBox(lb, fontheight, true);
    }


    /** 
     * Create a LabelBox for the given fontheight
     * @param    x horizontal base location in screen/pixel coordinates
     * @param    y vertical base location in screen/pixel coordinates
     * @author Tamara Munzner
     * @see	 AccordionDrawer.LabelBox
     */

    protected LabelBox makeLabelBox(int fontheight, int x, int y, float[] start, float[] end) {
		AccordionTreeDrawer d = (AccordionTreeDrawer)getCell().drawer;
//		if(d.drawDirection==AccordionDrawer.LEFT || d.drawDirection==AccordionDrawer.RIGHT)
		{
		  LabelBox lb=makeLabelBoxHor(fontheight, x,y, start, end);
		  return lb;
		}		
		//TODO: fix drawing in up and down directions
//        else if(d.drawDirection == AccordionDrawer.UP)
//		{
//			LabelBox lb = makeLabelBoxUp(fontheight, x, y);
//			return lb;
//		}
//		else
//		{
//			LabelBox lb = makeLabelBoxDown(fontheight, x, y);
//			return lb;
//				
//		}

       }

/**
 * make label box in horizontal direction
 * @param fontheight
 * @param x
 * @param y
 * @return
 */
	private LabelBox makeLabelBoxHor(int fontheight, int x, int y, float[] start, float[] end)
  {
	    AccordionTreeDrawer d = (AccordionTreeDrawer)getCell().drawer;
		int labelBottomLeft[] = new int[2];
		int labelTopRight[] = new int[2];
		int labelBottomLeftBG[] = new int[2];
//		String name = node.name; // make name here?
		String name = node.label;
		int namewidth = AccordionTreeDrawer.bff.stringWidth(name, fontheight);
		float ratio = 0.0f;
		int oldnamewidth = 0;
		int labelwidth = namewidth+2;
		int numChars;
		int[] winSize = {d.getWinsize(0), d.getWinsize(1)};
		if (namewidth >= winSize[0]) { //no hope of fitting, must truncate
			ratio = (winSize[0]-20.0f)/namewidth; 
			int namelength = name.length();
			int newlength = (int)(namelength*ratio);
			if (newlength < namelength)
				newlength = namelength;
			String newname = name.substring(namelength-newlength, namelength);
			name = newname;
			oldnamewidth = namewidth;
			namewidth = AccordionTreeDrawer.bff.stringWidth(name, fontheight);
			labelwidth = namewidth+2;
			labelTopRight[0] = winSize[0]-1;
		} 
		else {
			int moveover = (node.isLeaf()) ? labelwidth+d.labeloffset[0] 
			: -d.labeloffset[0];
			if(d.drawDirection != AccordionDrawer.RIGHT)
			{
				labelTopRight[0] = (d.labelposright) ?
				d.w2s(end[0],0)+moveover :
				d.w2s(start[0],0)-d.labeloffset[0];
			}
			if(d.drawDirection == AccordionDrawer.RIGHT){
				if(node.isLeaf())
				labelTopRight[0] = (d.labelposright) ?
			   moveover:
			   d.getWinsize(0)- d.w2s(start[0],0)+d.labeloffset[0];
			   else
				{
				labelTopRight[0] = (d.labelposright) ?
							d.getWinsize(0)-d.w2s(end[0],0)+labelwidth:
						   d.getWinsize(0)- d.w2s(start[0],0)+d.labeloffset[0];
				}
			}
		}
////	   cutting offscreen string when drawing in left direction, butterfly display	
//		 if(d.drawDirection == AccordionDrawer.RIGHT && labelTopRight[0] > d.w2s(1-posEnd[0], 0) && node.isLeaf())
//		   {
//				// ratio = d.w2s(1-posEnd[0], 0)/labelTopRight[0];
//				int namelength = d.w2s(1-posEnd[0], 0);
//				//int namelength =(int)( label.length()*ratio);
//				//int charlength = labelwidth/label.length();
//				int charwidth=0;
//				int numchar=0;
//		        int temp=0;
//				for(int i=0; charwidth<labelwidth-namelength;i++)
//				{
//					charwidth+=AccordionTreeDrawer.bff.charWidth(name,fontheight,i);
//					//temp = AccordionTreeDrawer.bff.stringWidth(name.substring(0,i), fontheight);
//					numchar++;
//				}		
////				  int temp = label.length() - namelength;
//                //charwidth=(int)(namewidth/name.length());
//                //numchar=(int)(labelTopRight[0]-namelength)/charwidth;
//				if(numchar >=name.length()-1) numchar = name.length()-1;
//				String newname=name.substring(numchar);
//				
//				name = newname;
//				namewidth = AccordionTreeDrawer.bff.stringWidth(name, fontheight);
//				labelwidth = namewidth;
//				labelTopRight[0] = d.w2s(1-posEnd[0], 0)-2 ;
//			   // labelBottomLeftBG[0] = labelTopRight[0] - labelwidth;
//			   labelBottomLeftBG[0] = 0;
//			}
		// either put label right and above junction
		// or put it left, vertically centered, and right-justified
		if (!d.labelposright)
		{
//			if(d.leftRight == 0 || d.leftRight == 1)
			labelTopRight[1] = d.w2s(start[1],1)-d.labeloffset[1]-fontheight;
//		   if(d.leftRight == 2)
//		   labelTopRight[1] = d.w2s(1-posStart[0],1)-d.labeloffset[1]-fontheight;
//		   if(d.leftRight == 3)
//			labelTopRight[1] = d.w2s(posStart[0],1)-d.labeloffset[1]-fontheight;
		}
		else 				
			labelTopRight[1] = (node.isLeaf()) ? 
			d.w2s(end[1],1) - (int)(fontheight/2.0) : 
			d.w2s(end[1],1)-d.labeloffset[1] - fontheight;       		
		
		// we pad background label box by one pixel on each side so
		// that text is never right up against the edge. height
		// already has enough padding
		int labelheight = fontheight+2;	
		labelBottomLeftBG[0] = labelTopRight[0] - labelwidth;
		labelBottomLeftBG[1] = labelTopRight[1] + labelheight;
		
	
		if (x>=0 && y>=0 && d.labelpopup) {
			labelBottomLeftBG[0] = x-(int)(labelwidth/2.0);
			labelTopRight[0] = labelBottomLeftBG[0]+labelwidth;
			labelBottomLeftBG[1] = y-5;
			labelTopRight[1] = labelBottomLeftBG[1]-labelheight;
		} 

		// if we would start offscreen, nudge over until it's all
		// onscreen. when the origin of the text string is offscreen,
		// opengl won't draw anything (sigh). we could truncate the
		// label text, but it seems more useful to just move the whole
		// thing over. while we're at it, make sure labels are always
		// totally onscreen.

		if ((labelBottomLeftBG[0]) < 0) {
			if(d.drawDirection != AccordionDrawer.RIGHT)
			{
		   labelBottomLeftBG[0] = 0;
		   labelTopRight[0] = labelBottomLeftBG[0] + labelwidth;
			}
		   if(d.drawDirection==AccordionDrawer.RIGHT)
		   {
		   labelTopRight[0] = (labelBottomLeftBG[0]+labelwidth > d.w2s(1-end[0],0)) ?
										 d.w2s(1-end[0], 0): labelBottomLeftBG[0] + labelwidth;
		   }
		} else if (labelTopRight[0] > winSize[0]) {	
			// do nothing - in leaf case, we want to truncate not nudge
			// as per hillis feedback.
		} 
//		System.out.println("top right is less than posend  ############"+labelTopRight[0]);
//		 System.out.println("pos end  ############" + d.w2s(1-posEnd[0], 0));
	   if(d.drawDirection!=AccordionDrawer.LEFT && labelTopRight[0] < d.w2s(1-end[0],0))
		{	
		  labelTopRight[0] = d.w2s(1-end[0],0)-1;
		  labelBottomLeftBG[0] = labelTopRight[0] - labelwidth;
		}
	
		if (labelBottomLeftBG[1] > winSize[1]) {
			labelBottomLeftBG[1] = winSize[1];
			labelTopRight[1] = labelBottomLeftBG[1]-labelheight;
		} else if (labelTopRight[1] < 0) {
			labelTopRight[1] = 0;
			labelBottomLeftBG[1] = labelheight;
		} 
		
		if (false) {
			System.out.println(name);
			System.out.println("tr " + labelTopRight[0] + " " + labelTopRight[1] + " blbg " + labelBottomLeftBG[0] + " " + labelBottomLeftBG[1] + " h " + labelheight + " asc " + AccordionTreeDrawer.bff.getAscent(fontheight) + " desc " + AccordionTreeDrawer.bff.getDescent(fontheight));
			System.out.println(" oldw " + oldnamewidth + " neww " + namewidth + " ratio + " + ratio + " ws " + winSize[0]);
		}

		// labelbuffer is used for occlusion computations to control
		// label density, but we don't want to draw the bg box that
		// big because it would look silly.

		labelBottomLeft[1] = labelBottomLeftBG[1] + d.labelbuffer[1];
		labelBottomLeft[0] = labelBottomLeftBG[0] - d.labelbuffer[0];
//		node.lb = 
		return new LabelBox( labelBottomLeft, labelTopRight, labelBottomLeftBG, d.getFrameNum(), name);
//		return node.lb;
  }


//TODO: implement drawing label boxes differently, so they work efficiently
//
///**
// * make label box in vertical direction
// * @param fontheight
// * @param x
// * @param y
// * @return
// */
//	private LabelBox makeLabelBoxUp(int fontheight, int x, int y) {
//	AccordionTreeDrawer d = (AccordionTreeDrawer)getCell().drawer;
//	int labelBottomLeft[] = new int[2];
//	int labelTopLeft[] = new int[2];
//	int labelTopRight[] = new int[2];
//	int labelBottomRight[] = new int[2];
//	int labelBottomLeftBG[] = new int[2];
//	int drawVert = 1;
//
////	String name = node.name; // make name here?
//	String name = node.label;
//	int namewidth = AccordionTreeDrawer.bff.stringWidth(name, fontheight);
////	int charwidth = (int)namewidth/name.length();
////	int charheight = fontheight;
//	float ratio = 0.0f;
//	int oldnamewidth = 0;
//	int labelwidth = namewidth+2;
//	int labelheight = fontheight+2;
//	int numChars;
//	int[] winSize = {d.getWinsize(0), d.getWinsize(1)};
//    float temp;  
//    temp = posEnd[0];
//    posEnd[0] = posEnd[1];
//    posEnd[1] = 1- temp;
//    temp = posStart[0];
//    posStart[0]= posStart[1];
//    posStart[1] = 1-temp;
//   
////	int moveover = (node.isLeaf()) ? labelwidth+d.labeloffset[0] 
////		: -d.labeloffset[0];
//	
//	// either put label right and above junction
//	// or put it left, vertically centered, and right-justified
//	if (!d.labelposright)
//	{
////		if(d.leftRight == 0 || d.leftRight == 1)
//		labelTopRight[1] = d.w2s(posStart[1],1)-d.labeloffset[1]-fontheight;
////	   if(d.leftRight == 2)
////	   labelTopRight[1] = d.w2s(1-posStart[0],1)-d.labeloffset[1]-fontheight;
////	   if(d.leftRight == 3)
////		labelTopRight[1] = d.w2s(posStart[0],1)-d.labeloffset[1]-fontheight;
//	}
//	else 		
//	{
//		labelBottomLeft[0] = (node.isLeaf()) ? 
//		d.w2s(posEnd[0],0) - (int)(fontheight/2.0) : 
//		d.w2s(posEnd[0],0)-d.labeloffset[0] - fontheight;
////		labelBottomRight[0] = labelBottomLeft[0] + labelheight;
//	}
//	// we pad background label box by one pixel on each side so
//	// that text is never right up against the edge. height
//	// already has enough padding
//
//	
//		labelBottomLeftBG[0] = d.w2s(posEnd[0], 0);
//		labelBottomLeftBG[1] = d.w2s(posEnd[1], 1);
//		labelTopRight[0] = labelBottomLeftBG[0] + labelheight;
//		labelTopRight[1] = labelBottomLeftBG[1] - labelwidth;
//	
//	if (x>=0 && y>=0 && d.labelpopup) {
//		labelBottomLeftBG[0] = x-(int)(labelwidth/2.0);
//			labelTopRight[0] = labelBottomLeftBG[0]+labelwidth;
//			labelBottomLeftBG[1] = y-5;
//			labelTopRight[1] = labelBottomLeftBG[1]-labelheight;
//	} 
//
//	// if we would start offscreen, nudge over until it's all
//	// onscreen. when the origin of the text string is offscreen,
//	// opengl won't draw anything (sigh). we could truncate the
//	// label text, but it seems more useful to just move the whole
//	// thing over. while we're at it, make sure labels are always
//	// totally onscreen.
//
//	if ((labelBottomLeftBG[0]) < 0) {
////	   labelBottomLeftBG[0] = 0;
////	   labelTopRight[0] = labelBottomLeftBG[0] + labelheight;
//	  
//	} else if (labelTopRight[0] > winSize[0]) {	
//		// do nothing - in leaf case, we want to truncate not nudge
//		// as per hillis feedback.
//	} 
//
//	if (labelBottomLeftBG[1] > winSize[1]) {
////		labelBottomLeftBG[1] = winSize[1];
////		labelTopRight[1] = labelBottomLeftBG[1]-labelwidth;
//	} else if (labelTopRight[1] < 0) {
////		labelTopRight[1] = 0;
////		labelBottomLeftBG[1] = labelTopRight[1]+labelwidth;
//	} 
//	
//	
//	// labelbuffer is used for occlusion computations to control
//	// label density, but we don't want to draw the bg box that
//	// big because it would look silly.
//
//	labelBottomLeft[1] = labelBottomLeftBG[1] + d.labelbuffer[1];
//	labelBottomLeft[0] = labelBottomLeftBG[0] - d.labelbuffer[0];
//
//    if(d.getDrawAngle() < 25.0)
//    {
//	float w = labelBottomLeft[1]*(float)Math.cos(Math.toRadians(d.getDrawAngle()));
//	float h = labelBottomLeft[1]*(float)Math.sin(Math.toRadians(d.getDrawAngle()));
//	labelTopRight[0] = labelTopRight[0] + (int)w;
//	labelTopRight[1] = labelBottomLeft[1] - (int)h; 	
////	labelTopRight[1] =labelBottomLeft[1] - labelBottomLeft[1]*(int)h;
////	labelTopRight[0] = labelBottomLeft[0] + labelBottomLeft[1]*(int)w;
//	if(labelTopRight[1]<0) labelTopRight[1] =0;
//    int cwidth = AccordionTreeDrawer.bff.charWidth(name,fontheight,0);
//	//labelBottomRight[0] = labelBottomLeft[0] + labelheight;
//	labelBottomRight[0] = labelBottomLeft[0] + cwidth;
//	labelBottomRight[1] = labelBottomLeft[1] ;
////	float newTR = labelTopRight[1]* java.lang.Math.sin(java.lang.Math.toRadians(30.0));
//	
//	labelTopLeft[1] = labelTopRight[1];
//	labelTopLeft[0] = labelTopRight[0]-labelheight;
//	lb = new  LabelBox(labelBottomLeft,  labelBottomRight, labelTopLeft, labelTopRight ,labelBottomLeftBG, d.getFrameNum(),  name, labelheight, d.getDrawAngle(), drawVert);
//	    }
//    else
//	lb = new LabelBox( labelBottomLeft, labelTopRight, labelBottomLeftBG, d.getFrameNum(), name);
//
// 	return lb;
//	}
//
//
//	/**
//	 * make label box in vertical direction
//	 * @param fontheight
//	 * @param x
//	 * @param y
//	 * @return
//	 */
//		private LabelBox makeLabelBoxDown(int fontheight, int x, int y) {
//
//    AccordionTreeDrawer d = (AccordionTreeDrawer)getCell().drawer;
//	int labelBottomLeft[] = new int[2];
//	int labelTopLeft[] = new int[2];
//	int labelTopRight[] = new int[2];
//	int labelBottomRight[] = new int[2];
//	int labelBottomLeftBG[] = new int[2];
//	int drawVert = 1;
//
////	String name = node.name; // make name here?
//	String name = node.label;
//	int namewidth = AccordionTreeDrawer.bff.stringWidth(name, fontheight);
////	int charwidth = (int)namewidth/name.length();
////	int charheight = fontheight;
//	float ratio = 0.0f;
//	int oldnamewidth = 0;
//	int labelwidth = namewidth+2;
//	int labelheight = fontheight+2;
//	int numChars;
//	int[] winSize = {d.getWinsize(0), d.getWinsize(1)};
//	float temp;  
//	temp = posEnd[1];
//	posEnd[1] = posEnd[0];
//	posEnd[0] = 1- temp;
//	temp = posStart[1];
//	posStart[1]= posStart[0];
//	posStart[0] = 1-temp;
////	int moveover = (node.isLeaf()) ? labelwidth+d.labeloffset[0] 
////		: -d.labeloffset[0];
//	
//	// either put label right and above junction
//	// or put it left, vertically centered, and right-justified
//	if (!d.labelposright)
//	{
////		if(d.leftRight == 0 || d.leftRight == 1)
//		labelTopRight[1] = d.w2s(posStart[1],1)-d.labeloffset[1]-fontheight;
////	   if(d.leftRight == 2)
////	   labelTopRight[1] = d.w2s(1-posStart[0],1)-d.labeloffset[1]-fontheight;
////	   if(d.leftRight == 3)
////		labelTopRight[1] = d.w2s(posStart[0],1)-d.labeloffset[1]-fontheight;
//	}
//	else 		
//	{
//		labelBottomLeft[0] = (node.isLeaf()) ? 
//		d.w2s(posEnd[0],0) - (int)(fontheight/2.0) : 
//		d.w2s(posEnd[0],0)-d.labeloffset[0] - fontheight;
////		labelBottomRight[0] = labelBottomLeft[0] + labelheight;
//	}
//	// we pad background label box by one pixel on each side so
//	// that text is never right up against the edge. height
//	// already has enough padding
//
//	
//		labelBottomLeftBG[0] = d.w2s(posEnd[0], 0);
//		labelBottomLeftBG[1] = d.w2s(posEnd[1], 1);
//		labelTopRight[0] = labelBottomLeftBG[0] + labelheight;
//		labelTopRight[1] = labelBottomLeftBG[1] - labelwidth;
//	
//	if (x>=0 && y>=0 && d.labelpopup) {
//		labelBottomLeftBG[0] = x-(int)(labelwidth/2.0);
//			labelTopRight[0] = labelBottomLeftBG[0]+labelwidth;
//			labelBottomLeftBG[1] = y-5;
//			labelTopRight[1] = labelBottomLeftBG[1]-labelheight;
//	} 
//
//	// if we would start offscreen, nudge over until it's all
//	// onscreen. when the origin of the text string is offscreen,
//	// opengl won't draw anything (sigh). we could truncate the
//	// label text, but it seems more useful to just move the whole
//	// thing over. while we're at it, make sure labels are always
//	// totally onscreen.
//
//	if ((labelBottomLeftBG[0]) < 0) {
////	   labelBottomLeftBG[0] = 0;
////	   labelTopRight[0] = labelBottomLeftBG[0] + labelheight;
//	  
//	} else if (labelTopRight[0] > winSize[0]) {	
//		// do nothing - in leaf case, we want to truncate not nudge
//		// as per hillis feedback.
//	} 
//
//	if (labelBottomLeftBG[1] > winSize[1]) {
////		labelBottomLeftBG[1] = winSize[1];
////		labelTopRight[1] = labelBottomLeftBG[1]-labelwidth;
//	} else if (labelTopRight[1] < 0) {
////		labelTopRight[1] = 0;
////		labelBottomLeftBG[1] = labelTopRight[1]+labelwidth;
//	} 
//	
//	
//	// labelbuffer is used for occlusion computations to control
//	// label density, but we don't want to draw the bg box that
//	// big because it would look silly.
//
//	labelBottomLeft[1] = labelBottomLeftBG[1] + d.labelbuffer[1];
//	labelBottomLeft[0] = labelBottomLeftBG[0] - d.labelbuffer[0];
//
//	if(d.getDrawAngle() < 25.0)
//	{
//	float w = labelBottomLeft[1]*(float)Math.cos(Math.toRadians(d.getDrawAngle()));
//	float h = labelBottomLeft[1]*(float)Math.sin(Math.toRadians(d.getDrawAngle()));
//	labelTopRight[0] = labelTopRight[0] + (int)w;
//	labelTopRight[1] = labelBottomLeft[1] - (int)h; 	
////	labelTopRight[1] =labelBottomLeft[1] - labelBottomLeft[1]*(int)h;
////	labelTopRight[0] = labelBottomLeft[0] + labelBottomLeft[1]*(int)w;
//	if(labelTopRight[1]<0) labelTopRight[1] =0;
//	int cwidth = AccordionTreeDrawer.bff.charWidth(name,fontheight,0);
//	//labelBottomRight[0] = labelBottomLeft[0] + labelheight;
//	labelBottomRight[0] = labelBottomLeft[0] + cwidth;
//	labelBottomRight[1] = labelBottomLeft[1] ;
////	float newTR = labelTopRight[1]* java.lang.Math.sin(java.lang.Math.toRadians(30.0));
//	
//	labelTopLeft[1] = labelTopRight[1];
//	labelTopLeft[0] = labelTopRight[0]-labelheight;
////	labelBottomLeft[1] = winSize[1] - labelBottomLeft[1];
////	labelBottomRight[1] = winSize[1] - labelBottomRight[1];
////	labelTopLeft[1] = winSize[1] - labelTopLeft[1];
////	labelTopRight[1] = winSize[1]  - labelTopRight[1];
//
//	lb = new LabelBox(labelBottomLeft,  labelBottomRight, labelTopLeft, labelTopRight ,labelBottomLeftBG, d.getFrameNum(),  name, labelheight, d.getDrawAngle(), drawVert);
//		}
//	else
//	lb = new LabelBox( labelBottomLeft, labelTopRight, labelBottomLeftBG, d.getFrameNum(), name);
//
//	return lb;
//
//				}


    /** 
     * Occlusion check of LabelBox against array of all drawn labels. 
     * {@link #drawnArray AccordionDrawer.drawnArray} is sorted in y -
     * the skinny direction, to avoid work. We do a binary search to
     * check for occlusions, so it's log in the number of pixels (with
     * a very small constant).
     * 
     * @author  Tamara Munzner
     * @see AccordionDrawer.LabelBox
     */
    private boolean intersectLabelBox(LabelBox lb) {
	AccordionTreeDrawer d = (AccordionTreeDrawer)getCell().drawer;

	// the list is sorted by y value. use binary search to see
	// if it will fit instead of checking against every item. 
	// find an item that has y overlap. check down from there
	// until no overlap, then check up until no overlap. stop
	// as soon as there's overlap, of course.
	boolean intersects = false;
	ArrayList drawnLabels = d.getDrawnLabels();
	int listlength = drawnLabels.size();
	//int high = listlength-1;
	int high=(listlength==1)?listlength: listlength-1;
	int low = 0;
	int current = -1;
	int prevcurrent = -2;
	int foundIndex = -1;
	LabelBox found;
	//while ((low < high) && (current != prevcurrent) && (!intersects)) {
	while ((low < high) && (!intersects) && (current<=high) && (foundIndex < 0))
	{
	    prevcurrent = current;
	    current = (int)((high+low)/2.0);
	    if (current == prevcurrent && low+1 == high) {
			low=high;
			current = high;
	    }
		if(current == listlength) current = current-1;
	    LabelBox prevlab = (LabelBox) drawnLabels.get(current);
	    //System.out.println("lb bl " + lb.bl[1] + " p tr " + prevlab.tr[1] + " lb tr " + lb.tr[1] + " p bl " + prevlab.bl[1]);
	    if (lb.compareBtoT(prevlab, low, high)) high = current;
	    else if (prevlab.compareBtoT(lb, low, high)) low = current;
	    else { // check intersections
		found = prevlab;
		foundIndex = current;
		//System.out.println("found: " + foundIndex);
		// we're doing a bit more work than necessary by
		// checking against prevlab from last iteration,
		// but it's the easiest way to ensure corner cases
		// work
		while (!intersects && prevlab.bottomBiggerThanTop(lb) && current >= 0) { 
		    // check down from found
		    // i.e. index in sorted list gets smaller, so
		    // y values get smaller also
		    prevlab = (LabelBox) drawnLabels.get(current);
		    intersects = lb.intersectBoxes(prevlab);
		    current--;
		    //System.out.println("cur-- now " + current);
		}
		//current = foundIndex+1;
		current = foundIndex;
		while (!intersects && lb.bottomBiggerThanTop(prevlab) && current <= listlength-1) { 
		    // check up from found
		    prevlab = (LabelBox) drawnLabels.get(current);
		    intersects = lb.intersectBoxes(prevlab);
		    current++;
		    //System.out.println("cur++ now " + current);
		}
	    }
	    //System.out.println("low " + low + " high " + high + " cur " + current + " prevcur " + prevcurrent + " intersects " + intersects);
 	}
	return intersects;
    }

    /** Draw a LabelBox. 
     * @author	 Tamara Munzner, Li Zhang
     * @param	 lb LabelBox to use
     * @param    fontheight Size of font to use, in points/pixels.
     * @param	 drawBig whether to draw maximum size ignoring occlusions
     */
    private void drawLabelBox(LabelBox lb, int fontheight, boolean drawBig) 
    {
		AccordionTreeDrawer d = (AccordionTreeDrawer)getCell().drawer;
		GLFunc gl = (GLFunc) d.getGraphicsStuff();
		ArrayList drawnLabels = d.getDrawnLabels();
		
		if (!drawBig) 
		{
	    	drawnLabels.add(lb);
	    	Collections.sort(drawnLabels);
		}
	
   		if(d.drawDirection == AccordionDrawer.LEFT || d.drawDirection == AccordionDrawer.RIGHT)
   			drawLabelBoxLR(d, gl, lb, fontheight, drawBig);
//   		else if(d.drawDirection==AccordionDrawer.UP)
//   			drawLabelBoxUp(d, gl, lb, fontheight, drawBig);
//    	else
//			drawLabelBoxDown(d, gl, lb, fontheight, drawBig);
    } 
    
/**
 * draw label box in left or right direction
 * @param lb
 * @param fontheight
 * @param drawBig
 */
	private void drawLabelBoxLR(AccordionTreeDrawer d,GLFunc gl, LabelBox lb, int fontheight, boolean drawBig)
	{
	    int[] pos = {lb.pos(0), lb.pos(1)};
		int[] topRightPos = {lb.topRightPos(0), lb.topRightPos(1)};
		
		
		if (drawBig) { // save off the pixels we're about to overwrite

			d.drawnlabelsize[0] = lb.size(0);
			d.drawnlabelsize[1] = lb.size(1);
			d.drawnlabelpos[0] = d.s2w(pos[0],0);
			d.drawnlabelpos[1] = d.s2w(pos[1],1);
			d.pixels = new float[d.drawnlabelsize[0]*3*d.drawnlabelsize[1]*3];
			
			// read pixels from frontbuffer
			gl.glReadBuffer(GLEnum.GL_FRONT);
			gl.glReadPixels(lb.pos(0),d.getWinsize(1) - lb.pos(1),d.drawnlabelsize[0],d.drawnlabelsize[1], GLEnum.GL_RGB, GLEnum.GL_FLOAT, d.pixels);
			d.flag = true; // first frame flag
			
		}

		
		double labelplane = d.getLabelplane();
		if (d.labeldrawback || drawBig) {
			float thecol[] = new float[3];
			if (drawBig) d.getLabelBackHiColor().getRGBColorComponents(thecol);
			else d.getLabelBackColor().getRGBColorComponents(thecol);
			if (!drawBig && d.labeltransp) {
			gl.glColor4f(thecol[0], thecol[1], thecol[2], .5f);
			
			
				
			} else {
			gl.glColor3f(thecol[0], thecol[1], thecol[2]);
			}
	        if(d.drawDirection!=AccordionDrawer.RIGHT)
	        {
	        		// change glstate locally
					//gl.glDisable(GLEnum.GL_DEPTH_TEST);		
					//gl.glColor3f(1.0f,1.0f,1.0f);
					gl.glBegin(GLEnum.GL_POLYGON);
					gl.glVertex3d(d.s2w(pos[0], 0), d.s2w(pos[1], 1), labelplane);
					gl.glVertex3d(d.s2w(topRightPos[0], 0), d.s2w(pos[1], 1), labelplane);
					gl.glVertex3d(d.s2w(topRightPos[0], 0), d.s2w(topRightPos[1], 1), labelplane);
					gl.glVertex3d(d.s2w(pos[0], 0), d.s2w(topRightPos[1], 1), labelplane);
					gl.glEnd();
					//gl.glEnable(GLEnum.GL_DEPTH_TEST);
					
						
	          }
			if(d.drawDirection==AccordionDrawer.RIGHT)
		   {
				//gl.glDisable(GLEnum.GL_DEPTH_TEST);
			   	gl.glBegin(GLEnum.GL_POLYGON);
			   	gl.glVertex3d(d.s2w(pos[0],0), d.s2w(pos[1],1), labelplane);
			   	gl.glVertex3d(d.s2w(topRightPos[0],0), d.s2w(pos[1],1), labelplane);
		   		gl.glVertex3d(d.s2w(topRightPos[0],0), d.s2w(topRightPos[1],1), labelplane);
		   		gl.glVertex3d(d.s2w(pos[0],0), d.s2w(topRightPos[1],1), labelplane);
		   		gl.glEnd();
				//gl.glEnable(GLEnum.GL_DEPTH_TEST);
		   }
	     	
		}
		int descent = AccordionTreeDrawer.bff.getDescent(fontheight);
		if (drawBig) 
			d.setColorGL(d.getLabelHiColor());
		else {
			if (!d.labeldrawback) {
			// first draw 4X (nw/ne/sw/se) in outline color 
			d.setColorGL(d.getLabelBackColor());
			int bx = pos[0] + 1;
			int by = pos[1]-descent-1;
			}		
		}
			d.setColorGL(d.getLabelColor());
			
			
			if(d.drawDirection ==AccordionDrawer.LEFT || d.drawDirection ==AccordionDrawer.RIGHT)
			{
			gl.glRasterPos3d((float)d.s2w(pos[0] +1,0), (float)d.s2w(pos[1]-descent-1,1),   
				(float)labelplane);
				AccordionTreeDrawer.bff.drawString(gl, lb.getName(), fontheight);		
			}
			
			
			
	}
	
// TODO: is there a better way of drawing labels in different directions?
//	
//	
///**
// * draw label box in vertical direction
// * @param d
// * @param gl
// * @param lb
// * @param fontheight
// * @param drawBig
// */
//private void drawLabelBoxUp(AccordionTreeDrawer d,GLFunc gl, LabelBox lb, int fontheight, boolean drawBig)
//	{
//		int[] pos = {lb.pos(0), lb.pos(1)};       
//		int[] topRightPos = {lb.topRightPos(0), lb.topRightPos(1)};
//		//String name = lb.getName();
//		String name = getName();
//		int cwidth;
//        float[][] p = new float[name.length()][];
////		for(int i=0; i<name.length(); i++)
////		{
////		 cwidth= AccordionTreeDrawer.bff.charWidth(name,fontheight,i);
////		 xInterval[i] = cwidth;         
////		 float t =  cwidth*(float)java.lang.Math.tan(java.lang.Math.toRadians(d.getDrawAngle()));
////		 yInterval[i] = (int)t;
////		}
////		d.xInterval = xInterval;
////		d.yInterval = yInterval;
//		if (drawBig) { // save off the pixels we're about to overwrite
//
//			d.drawnlabelsize[0] =fontheight+6;
//			d.drawnlabelsize[1] =fontheight;
//			d.drawnlabelpos[0] = d.s2w(pos[0]-3,0);
//			d.drawnlabelpos[1] = d.s2w(pos[1],1);
//	        int x=pos[0]-3;
//	        int y=pos[1];
//	        for(int i = 0; i<name.length(); i++)
//	        {
//			d.pixels = new float[d.drawnlabelsize[0]*3*d.drawnlabelsize[1]*3];
//            p[i]=d.pixels;
//			gl.glReadBuffer(GLEnum.GL_FRONT);
//			gl.glReadPixels(x,d.getWinsize(1)-y,d.drawnlabelsize[0],d.drawnlabelsize[1], GLEnum.GL_RGB, GLEnum.GL_FLOAT, p[i]);
////			x+=xInterval[i];
////			y-=yInterval[i];
//	        }
//	        d.pal = p;
//	
//		}
//
//		
//		// padding of one pixel on right and bottom, and make sure
//		// we leave enough vertical room for character parts that
//		// descend below font baseline.
//		double labelplane = d.getLabelplane();
//		if (d.labeldrawback || drawBig) {
//			float thecol[] = new float[3];
//			if (drawBig) d.getLabelBackHiColor().getRGBColorComponents(thecol);
//			else d.getLabelBackColor().getRGBColorComponents(thecol);
//			if (!drawBig && d.labeltransp) {
//			gl.glColor4f(thecol[0], thecol[1], thecol[2], .5f);
//		
//		
//		
//			} else {
//			gl.glColor3f(thecol[0], thecol[1], thecol[2]);
//			}
//
//            int x=0;
//            int y=0;
//            int c;
//            for(int i=0; i<lb.getName().length(); i++)
//            {
//			
//			c=10;
//			gl.glDisable(GLEnum.GL_DEPTH_TEST);
//			gl.glBegin(GLEnum.GL_POLYGON);
//			gl.glVertex3d(d.s2w(pos[0]+x-3,0), d.s2w(pos[1]-y,1), labelplane);
//			gl.glVertex3d(d.s2w(pos[0]+x+c+3,0), d.s2w(pos[1]-y,1), labelplane);
//			gl.glVertex3d(d.s2w(pos[0]+x+c+3,0), d.s2w(pos[1]-y-fontheight,1), labelplane);
//			gl.glVertex3d(d.s2w(pos[0]+x-3,0), d.s2w(pos[1]-y-fontheight,1), labelplane);
//			gl.glEnd();
//			gl.glEnable(GLEnum.GL_DEPTH_TEST);
////			x+=xInterval[i];
////			y+=yInterval[i];
//            }
//
//		}
//		int descent = AccordionTreeDrawer.bff.getDescent(fontheight);
//		if (drawBig) 
//			d.setColorGL(d.getLabelHiColor());
//		else {
//			if (!d.labeldrawback) {
//			// first draw 4X (nw/ne/sw/se) in outline color 
//			d.setColorGL(d.getLabelBackColor());
//			int by = pos[1]-descent-1;
//			int bx = pos[0] ;
//	
//		}
//			d.setColorGL(d.getLabelColor());
//		}
//	
//			int x=0;
//			int y =descent+1 ;	
//			for(int i=0; i<name.length(); i++)
//			{				            
//				gl.glRasterPos3d((float)d.s2w(pos[0] +x,0), (float)d.s2w(pos[1]-y,1),   
//				(float)labelplane);
//				AccordionTreeDrawer.bff.drawString(gl, name.substring(i,i+1), fontheight);
//				
////                x+=xInterval[i];
////                y+=yInterval[i];
//			}
//	}
//	
//// draw in down direction
//	
//private void drawLabelBoxDown(AccordionTreeDrawer d,GLFunc gl, LabelBox lb, int fontheight, boolean drawBig)
//	{
//		int[] pos = {lb.pos(0), lb.pos(1)};
//		int[] topRightPos = {lb.topRightPos(0), lb.topRightPos(1)};
//		String name = lb.getName();
//		float[][] p = new float[name.length()][]; 		
//		int cwidth;
//
////		for(int i=0; i<name.length(); i++)
////		{
////			 cwidth= AccordionTreeDrawer.bff.charWidth(name,fontheight,i);
////			 xInterval[i] = cwidth;         
////			 float t =  cwidth*(float)Math.tan(java.lang.Math.toRadians(d.getDrawAngle()));
////		 	yInterval[i] = (int)t;
////		}
////		
////		d.xInterval = xInterval;
////		d.yInterval = yInterval;
//		if (drawBig) { // save off the pixels we're about to overwrite
//   
//			d.drawnlabelsize[0] =fontheight+6;
//			d.drawnlabelsize[1] =lb.size(0)+2;
//			d.drawnlabelpos[0] = d.s2w(pos[0]-3,0);
//			d.drawnlabelpos[1] = d.s2w(pos[1]+fontheight+4,1);
//			int x=pos[0]-3;
//			int y=pos[1]+fontheight+4;
//			for(int i = 0; i<lb.getName().length(); i++)
//			{
//				d.pixels = new float[d.drawnlabelsize[0]*3*d.drawnlabelsize[1]*3];
//
//    	    	p[i] = d.pixels;
//				gl.glReadBuffer(GLEnum.GL_FRONT);
//
//				gl.glReadPixels(x,d.getWinsize(1)-y,d.drawnlabelsize[0],d.drawnlabelsize[1], GLEnum.GL_RGB, GLEnum.GL_FLOAT, p[i]);
////				x+=xInterval[i];
////				y+=yInterval[i];
//				if(y>d.getWinsize(1)) y=d.getWinsize(1)-1;
//			}
//             d.pal = p;
//	
//		}
//
//		
//		// padding of one pixel on right and bottom, and make sure
//		// we leave enough vertical room for character parts that
//		// descend below font baseline.
//		double labelplane = d.getLabelplane();
//		if (d.labeldrawback || drawBig) {
//			float thecol[] = new float[3];
//			if (drawBig) d.getLabelBackHiColor().getRGBColorComponents(thecol);
//			else d.getLabelBackColor().getRGBColorComponents(thecol);
//			if (!drawBig && d.labeltransp) {
//			gl.glColor4f(thecol[0], thecol[1], thecol[2], .5f);
//			
//			
//			} else {
//			gl.glColor3f(thecol[0], thecol[1], thecol[2]);
//			}
//			int x=0;
//			int y=0;
//			int c;
//			for(int i=0; i<lb.getName().length(); i++)
//			{
//				//c=AccordionTreeDrawer.bff.charWidth(lb.getName(),fontheight,i);
//				c=10;
//				gl.glDisable(GLEnum.GL_DEPTH_TEST);
//				gl.glBegin(GLEnum.GL_POLYGON);
//				gl.glVertex3d(d.s2w(pos[0]+x-3,0), d.s2w(pos[1]+y,1), labelplane);
//				gl.glVertex3d(d.s2w(pos[0]+x+c+3,0), d.s2w(pos[1]+y,1), labelplane);
//				gl.glVertex3d(d.s2w(pos[0]+x+c+3,0), d.s2w(pos[1]+y+fontheight,1), labelplane);
//				gl.glVertex3d(d.s2w(pos[0]+x-3,0), d.s2w(pos[1]+y+fontheight,1), labelplane);
//				gl.glEnd();
//				gl.glEnable(GLEnum.GL_DEPTH_TEST);
////				x+=xInterval[i];
////				y+=yInterval[i];
//			}
//	
//		
//		}
//		int descent = AccordionTreeDrawer.bff.getDescent(fontheight);
//		if (drawBig) 
//			d.setColorGL(d.getLabelHiColor());
//		else {
//			if (!d.labeldrawback) {
//			// first draw 4X (nw/ne/sw/se) in outline color 
//			d.setColorGL(d.getLabelBackColor());
//		int by = pos[1]+descent+10;
//			int bx = pos[0] ;
//	
//
//		}
//			d.setColorGL(d.getLabelColor());
//		}
//	
//			int x=0;
//			int y =descent+1 ;	
//			for(int i=0; i<name.length(); i++)
//			{				            
//				gl.glRasterPos3d((float)d.s2w(pos[0] +x,0), (float)d.s2w(pos[1]+y+10,1),   
//				(float)labelplane);
//				AccordionTreeDrawer.bff.drawString(gl, name.substring(i,i+1), fontheight);
////				x+=xInterval[i];
////				y+=yInterval[i];
//			
//			}
//	}

// TODO: this is actually only required by treenode
//	/**
//	 * Determines whether the window coordinates (x,y) fall on this TreeEdge
//	 * There is a "snap" factor, determined by the pickFuzz field of
//	 * the enclosing cell.
//	 *
//	 * @author   Tamara Munzner
//	 * @see      AccordionDrawer.GridCell
//	 * @see      AccordionDrawer.AccordionTreeDrawer
//	 */			
//	// convert everything to window coords (0,1)
//	// x and y are initially pixels
//	public boolean pick(int x, int y)
//	{
//		GridCell c = getCell();
//		AccordionTreeDrawer d = (AccordionTreeDrawer)c.drawer;
//		if (horiz)
//		{ // horizontal line, check in Y direction
//			double pickFuzzAbsolute = d.s2w(d.pickFuzz, AccordionDrawer.Y);
//			double yAbsolute = d.s2w(y, AccordionDrawer.Y);
//			return yAbsolute > c.getMin(AccordionDrawer.Y) - pickFuzzAbsolute &&
//				yAbsolute < c.getMax(AccordionDrawer.Y) + pickFuzzAbsolute;
//		}
//		else
//		{
//			double pickFuzzAbsolute = d.s2w(d.pickFuzz, AccordionDrawer.X);
//			double xAbsolute = d.s2w(x, AccordionDrawer.X);
//			return xAbsolute > c.getMin(AccordionDrawer.X) - pickFuzzAbsolute &&
//				xAbsolute < c.getMax(AccordionDrawer.X) + pickFuzzAbsolute;
//
//		}
//	}


	public void close()
	{
	}
	

};

